/*
 * SUNTAC U-Cable P (PHS/PDC usb adapter) driver
 *
 *   Copyright (C) 2002-2003 Toshio Sato (CXL00145@nifty.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * (14/Jan/2003) Toshio Sato
 *    V1.4.0 release. kernel2.4.19 (or later) support.
 *    (kernel2.2 is not supported.)
 *
 * (07/Jul/2002) Toshio Sato
 *    v1.3.0 release. Added support for DS-96L (U-Cable type D).
 *
 * (04/Apr/2002) Toshio Sato
 *    v1.2.0 release. Known Bug Fix.
 *
 * (29/Mar/2002) Toshio Sato
 *    v1.1.1 release. Added usb_submit_urb() error check routine. Thanks to Mamedore.
 *
 * (12/Mar/2002) Toshio Sato
 *    v1.1.0 release. Support linux kernel 2.4.x.
 *
 * (08/Jan/2002) Toshio Sato
 *    v1.0.0 release. for linux kernel 2.2.x.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#ifdef CONFIG_USB_SERIAL_DEBUG
	static int debug = 1;
#else
	static int debug;
#endif

#include "usb-serial.h"
#include "suntac_ucablep.h"

/*
 * Version Information
 */
#define DRIVER_VERSION "v1.4.0"
#define DRIVER_AUTHOR "Toshio Sato <CXL00145@nifty.com>"
#define DRIVER_DESC "SUNTAC U-Cable P driver"

/*
 * Function prototypes
 */
static int  suntac_ucp_startup             (struct usb_serial *serial);
static void suntac_ucp_shutdown            (struct usb_serial *serial);
static int  suntac_ucp_open                (struct usb_serial_port *port, struct file *filp);
static void suntac_ucp_close               (struct usb_serial_port *port, struct file *filp);
static int  suntac_ucp_write               (struct usb_serial_port *port, int from_user, const unsigned char *buf, int count);
static void suntac_ucp_write_bulk_callback (struct urb *urb);
static void suntac_ucp_read_int_callback   (struct urb *urb);
static void suntac_ucp_read_bulk_callback  (struct urb *urb);
static void suntac_ucp_set_termios         (struct usb_serial_port *port, struct termios * old);
static int  suntac_ucp_ioctl               (struct usb_serial_port *port, struct file * file, unsigned int cmd, unsigned long arg);
static void suntac_ucp_break_ctl           (struct usb_serial_port *port, int break_state );
static void suntac_ucp_throttle            (struct usb_serial_port *port);
static void suntac_ucp_unthrottle          (struct usb_serial_port *port);
static int  suntac_ucp_set_protocol        (struct usb_serial *serial, unsigned char lcr);
static int  suntac_ucp_write_room          (struct usb_serial_port *port);
static int  suntac_ucp_chars_in_buffer     (struct usb_serial_port *port);


static struct usb_device_id suntac_ucp_id_table [] = {
	{ USB_DEVICE(SUNTAC_VID, SUNTAC_UCP_P_PID)  },
	{ USB_DEVICE(SUNTAC_VID, SUNTAC_UCP_D_PID)  },
	{ }					/* Terminating entry */
};

static __devinitdata struct usb_device_id id_table_combined [] = {
	{ USB_DEVICE(SUNTAC_VID, SUNTAC_UCP_P_PID)  },
	{ USB_DEVICE(SUNTAC_VID, SUNTAC_UCP_D_PID)  },
	{ }					/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, id_table_combined);


/*
 * All of the device info needed for the SUNTAC U-Cabel cable.
 */
static struct usb_serial_device_type suntac_ucp_device = {
	.owner =		THIS_MODULE,
	.name =			"SUNTAC U-Cable P",
	.id_table =		suntac_ucp_id_table,

	.num_interrupt_in =	1,
	.num_bulk_in =		1,
	.num_bulk_out =		1,
	.num_ports =		1,

	.open =			suntac_ucp_open,
	.close =		suntac_ucp_close,
	.throttle =		suntac_ucp_throttle,
	.unthrottle =		suntac_ucp_unthrottle,

	.startup =		suntac_ucp_startup,
	.shutdown =		suntac_ucp_shutdown,
	.ioctl =		suntac_ucp_ioctl,
	.set_termios =		suntac_ucp_set_termios,
	.write =		suntac_ucp_write,
	.write_room =		suntac_ucp_write_room,
	.chars_in_buffer =	suntac_ucp_chars_in_buffer,
	.write_bulk_callback =	suntac_ucp_write_bulk_callback,
	.read_bulk_callback =	suntac_ucp_read_bulk_callback,
	.read_int_callback =	suntac_ucp_read_int_callback,

	.break_ctl =		suntac_ucp_break_ctl,
};

/*
 *	Write urb pool
 */
#define	NUM_URBS			32
#define	URB_TRANSFER_BUFFER_SIZE	64

/*
 *	Interrupt Transfer Data Structure
 */
struct ucp_private {
	unsigned char	dev_state;		/* Device Status */
	unsigned char	line_state;		/* Line   Status */
};

/*
 *	Private Data Structure
 *
 *	(Device and Modem line status, urb_pool, etc.)
 */
struct suntac_ucp_private {
	struct ucp_private ucp;			/* interrupt transfer receive data */

	unsigned char	baudrate;		/* DeviceRegister: DTE Speed                 */
	unsigned char	ctrlline;		/* DeviceRegister: DTR,RTS,BREAK             */
	unsigned char	databits;		/* DeviceRegister: DATAbits, StopBit, PARITY */

	unsigned long	control_state;		/* Modem Line Setting (for TIOCM) */

	int		urb_read_index;		/* Next read index	*/
	int		urb_write_index;	/* Next write index	*/
	int		urb_free_cnt;		/* free urb count	*/
	spinlock_t	write_urb_pool_lock;
	unsigned char	writebuf[NUM_URBS][URB_TRANSFER_BUFFER_SIZE];
	unsigned char	writelen[NUM_URBS];
};

/*
 * Handle vendor specific USB requests
 */

#define WDR_TIMEOUT (HZ * 5 ) /* default urb timeout */


/*
 *	U-Cable P Initialize
 *	(115200bps, Data=8bit, Stop=Don't care, Non parity, Hardflow)
 */
static int suntac_ucp_deviceinit( struct usb_serial *serial )
{
	int		result = 0;
	unsigned char	*transfer_buffer = kmalloc( 256, GFP_KERNEL );

	if ( !transfer_buffer ) {
		err("%s - kmalloc(%d) failed.", __FUNCTION__, 256);
		result = -ENOMEM;
	}else{
		do {
			/* DeviceReset */
			result = usb_control_msg(
				serial->dev,
				usb_rcvctrlpipe(serial->dev, 0),
				REQ_DEVRESET,		/* Request     : setup_packet 2byte   */
				REQTYPE_DEVICE,		/* RequestType : setup_packet 1byte   */
				0x0,			/* Value       : setup_packet 3byte   */
				0x0,			/* Index       : setup_packet 4byte   */
				transfer_buffer,
				0x02,			/* length      : setup_packet 6-7byte */
				WDR_TIMEOUT);
			if (result < 0) {
				err("DeviceReset (error = %d)", result);
				break;
			}

			/* DeviceReset's return code : OK(01 00), NG(00 00) */
			if ( (transfer_buffer[0] != 0x01) || (transfer_buffer[1] != 0x00 ) ) {
				err("DeviceReset rc= %02x %02x", transfer_buffer[0], transfer_buffer[1] );
				result = -1;
				break;
			}

			/* BaudRateSet */
			result = usb_control_msg(
				serial->dev,
				usb_sndctrlpipe(serial->dev, 0),
				REQ_SET_BAUDRATE,	/* Request     : setup_packet 2byte   */
				REQTYPE_SERIAL,		/* RequestType : setup_packet 1byte   */
				0x0a,			/* Value       : setup_packet 3byte   */
				0x0,			/* Index       : setup_packet 4byte   */
				NULL,
				0x0,			/* length      : setup_packet 6-7byte */
				WDR_TIMEOUT);
			if (result < 0) {
				err("BaudRateSet (error = %d)", result);
			}

			/* LineSet */
			result = usb_control_msg(
				serial->dev,
				usb_sndctrlpipe(serial->dev, 0),
				REQ_SET_LINESTATE,	/* Request     : setup_packet 2byte   */
				REQTYPE_SERIAL,		/* RequestType : setup_packet 1byte   */
				0x0b,			/* Value       : setup_packet 3byte   */
				0x0,			/* Index       : setup_packet 4byte   */
				NULL,
				0x0,			/* length      : setup_packet 6-7byte */
				WDR_TIMEOUT);
			if (result < 0) {
				err("LineSet (error = %d)", result);
			}

			/* ProtocolSet */
			result = usb_control_msg(
				serial->dev,
				usb_sndctrlpipe(serial->dev, 0),
				REQ_SET_DATABITS,	/* Request     : setup_packet 2byte   */
				REQTYPE_SERIAL,		/* RequestType : setup_packet 1byte   */
				0x03,			/* Value       : setup_packet 3byte   */
				0x0,			/* Index       : setup_packet 4byte   */
				NULL,
				0x0,			/* length      : setup_packet 6-7byte */
				WDR_TIMEOUT);
			if (result < 0) {
				err("DataLenSet (error = %d)", result);
			}

			result = 0;
		} while(0);

		kfree( transfer_buffer );
	}

	return result;

} /* suntac_ucp_deviceinit */


/*
 *	Calcurate Baudrate Value for U-Cable P
 *	(default 115200 bps)
 */
static int suntac_ucp_calculate_baud_rate(struct usb_serial *serial, int value)
{
	switch( value ) {
		case    110: return 0x00;
		case    300: return 0x01;
		case    600: return 0x02;
		case   1200: return 0x03;
		case   2400: return 0x04;
		case   4800: return 0x05;
		case   9600: return 0x06;
		case  19200: return 0x07;
		case  38400: return 0x08;
		case  57600: return 0x09;
		case 115200: return 0x0a;
		case 230400: return 0x0b;
		case 460800: return 0x0c;
		case 921600: return 0x0d;
		default:     return 0x0a;	/* default 115200bps */
	}
}


/*
 *	Set serial speed.
 */
static int suntac_ucp_set_baud_rate(struct usb_serial *serial, int value)
{
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)serial->port->private;
	int	rc;

	/* calc value for baudrate */
	priv->baudrate = suntac_ucp_calculate_baud_rate(serial, value);

	/* BaudRateSet */
	rc = usb_control_msg(
		serial->dev,
		usb_sndctrlpipe(serial->dev, 0),
		REQ_SET_BAUDRATE,	/* Request     */
		REQTYPE_SERIAL,		/* RequestType */
		priv->baudrate,		/* Value       */
		0x0,			/* Index       */
		NULL,
		0x0,			/* length      */
		WDR_TIMEOUT);
	if (rc < 0) {
		err("Set BAUD RATE %d failed (error = %d)", value, rc);
	}

	dbg("set_baud_rate: value: %d, bRequest: %d", value, priv->baudrate );
	return rc;
} /* suntac_ucp_set_baud_rate */


/*
 *	Set DATAbits, PARITY, STOPbits
 */
static int suntac_ucp_set_protocol(struct usb_serial *serial, unsigned char protocol)
{
	int rc;

	/* ProtocolSet */
	rc = usb_control_msg(
		serial->dev,
		usb_sndctrlpipe(serial->dev, 0),
		REQ_SET_DATABITS,
		REQTYPE_SERIAL,
		protocol,
		0x0,
		NULL,
		0x0,
		WDR_TIMEOUT);
	if (rc < 0) {
		err("Set Protocol 0x%x failed (error = %d)", protocol, rc);
	}

	dbg("set_protocol: 0x%x", protocol);
	return rc;
} /* suntac_ucp_set_protocol */


/*
 *	Set RTS, DTR, BREAK
 */
static int suntac_ucp_mdmctrl_direct(struct usb_serial *serial, unsigned char lcr)
{
	int rc;

	/* LineSet */
	rc = usb_control_msg(
		serial->dev,
		usb_sndctrlpipe(serial->dev, 0),
		REQ_SET_LINESTATE,
		REQTYPE_SERIAL,
		lcr,
		0x0,
		NULL,
		0x0,
		WDR_TIMEOUT);
	if (rc < 0) {
		err("Set MODEM CTRL Direct 0x%x failed (error = %d)", lcr, rc);
	}

	dbg("mdmctrl_direct: 0x%x", lcr);
	return rc;
} /* suntac_ucp_mdmctrl_direct */


/*
 *	Control DTR, RTS from control_state.
 */
static int suntac_ucp_mdmctl(struct usb_serial *serial,
				   unsigned long control_state)
{
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)serial->port->private;
	int rc;

	if (control_state & TIOCM_DTR)
		priv->ctrlline |= SUNTAC_UCP_DTR_bit;
	else
		priv->ctrlline &= ~SUNTAC_UCP_DTR_bit;

	if (control_state & TIOCM_RTS)
		priv->ctrlline |= SUNTAC_UCP_RTS_bit;
	else
		priv->ctrlline &= ~SUNTAC_UCP_RTS_bit;

	rc = suntac_ucp_mdmctrl_direct( serial, priv->ctrlline );

	if (rc < 0)
		err("Set MODEM CTRL 0x%x failed (error = %d)", priv->ctrlline, rc);

	dbg("mdmctl: state=0x%lx ==> mcr=0x%x", control_state, priv->ctrlline);

	return rc;
} /* suntac_ucp_mdmctl */


/*
 *	Make control_state from line_state.
 */
static void suntac_ucp_msr_to_state(unsigned long *control_state, unsigned char msr)
{
	if (msr & SUNTAC_UCP_MSR_DSR)
		*control_state |=  TIOCM_DSR;
	else
		*control_state &= ~TIOCM_DSR;

	if (msr & SUNTAC_UCP_MSR_CTS)
		*control_state |=  TIOCM_CTS;
	else
		*control_state &= ~TIOCM_CTS;

	if (msr & SUNTAC_UCP_MSR_RI)
		*control_state |=  TIOCM_RI;
	else
		*control_state &= ~TIOCM_RI;

	if (msr & SUNTAC_UCP_MSR_CD)
		*control_state |=  TIOCM_CD;
	else
		*control_state &= ~TIOCM_CD;

	dbg("msr_to_state: msr=0x%x ==> state=0x%lx", msr, *control_state);

} /* suntac_ucp_msr_to_state */


/*
 * Driver's tty interface functions
 */


/*
 * When USB-Device was connected, this routine is kicked.
 */
static int suntac_ucp_startup (struct usb_serial *serial)
{
	struct suntac_ucp_private	*priv;

	/* allocate the private data structure */
	serial->port->private = kmalloc(sizeof(struct suntac_ucp_private), GFP_KERNEL);

	if (!serial->port->private) {
		err("%s - kmalloc(%d) failed.", __FUNCTION__, 256);
		return -ENOMEM;
	}

	priv = (struct suntac_ucp_private *)serial->port->private;

	/* set initial values for control structures */
	priv->control_state = 0;
	priv->ucp.dev_state = 0;
	priv->ucp.line_state = 0;
	priv->baudrate = 0x0a;	/* 115200bps */
	priv->ctrlline = 0x0b;	/* DTR, RTS */
	priv->databits = 0x03;	/* data8bit, stop1bit, nonParity */

	priv->urb_read_index	= 0;		/* initialize read_index */
	priv->urb_write_index	= 0;		/* initialize write_index */
	priv->urb_free_cnt	= NUM_URBS;	/* free urb counts */

	/* create our write urb pool and transfer buffers */
	spin_lock_init( &priv->write_urb_pool_lock );

	init_waitqueue_head(&serial->port->write_wait);
	
	return (0);
} /* suntac_ucp_startup */


/*
 * When USB-Device was dis-connected, this routine is kicked.
 */
static void suntac_ucp_shutdown (struct usb_serial *serial)
{
	int	i;
	
	dbg("%s", __FUNCTION__);

	/* stop reads and writes on all ports */
	for (i=0; i < serial->num_ports; ++i) {
		/* My special items, the standard routines free my urbs */
		if ( serial->port[i].private != NULL ) {
			kfree(serial->port[i].private);
			serial->port[i].private = NULL;
		}
	}
} /* suntac_ucp_shutdown */


/*
 *	Port open and Device initialize & interrupt urb start.
 *
 *	(transfer buffer size change to 192byte).
 */
static int  suntac_ucp_open (struct usb_serial_port *port, struct file *filp)
{
	struct usb_serial *serial = port->serial;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;
	int result = 0;

	if (port_paranoia_check( port, __FUNCTION__ ))
		return -ENODEV;

	dbg("%s - port %d", __FUNCTION__, port->number);

	/* transfer buffer size change ( 64 -> 192 ) */

	if (port->bulk_in_buffer)
		kfree(port->bulk_in_buffer);

	port->bulk_in_buffer = kmalloc( 64 * 3, GFP_KERNEL );
	if ( ! port->bulk_in_buffer ) {
		err(" Couldn't allocate bulk_in_buffer");
		result = -ENOMEM;
		goto open_error;
	}
	port->read_urb->transfer_buffer = port->bulk_in_buffer;
	port->read_urb->transfer_buffer_length = 64 * 3;

	/* set initial values for control structures */

	priv->control_state  = TIOCM_RTS | TIOCM_DTR;
	priv->ucp.dev_state  = 0;
	priv->ucp.line_state = 0;
	priv->baudrate       = 0x0a;	/* 115200bps */
	priv->ctrlline       = 0x0b;	/* DTR, RTS On */
	priv->databits       = 0x03;	/* Data8bit, Stop1bit, NonParity */

	suntac_ucp_msr_to_state(&priv->control_state, priv->ucp.line_state);

	/* U-Cable P Initialize */
	result = suntac_ucp_deviceinit( serial );
	if ( result ) {
		err(" Couldn't DeviceReset U-Cable P");
		goto open_error;
	}

	if ( port->tty )
		port->tty->low_latency = 1;

	/*
		Interrupt Transfer Start.
	*/
	port->interrupt_in_urb->dev = port->serial->dev;

	port->interrupt_in_urb->interval = 1;
	port->interrupt_in_urb->transfer_buffer_length = 2;	/* 2byte */

	result = usb_submit_urb(port->interrupt_in_urb);
	if ( result ) {
		err(" usb_submit_urb(read int) failed");
		goto open_error;
	}

open_error:
	return result;

} /* suntac_ucp_open */


/*
 *	Port close and Device close.
 *	(All urb unlink)
 */
static void suntac_ucp_close (struct usb_serial_port *port, struct file *filp)
{
	struct usb_serial *serial;
	int rc;

	if (port_paranoia_check (port, __FUNCTION__))
		return;

	dbg("%s - port %d", __FUNCTION__, port->number);

	serial = get_usb_serial(port, __FUNCTION__);
	if ( !serial )
		return;

	if (serial->dev) {
		/* shutdown all urb */
		dbg("unlink write_urb");
		usb_unlink_urb (port->write_urb);

		dbg("unlink read_urb");
		usb_unlink_urb (port->read_urb);

		dbg("unlink interrupt_in_urb");
		usb_unlink_urb (port->interrupt_in_urb);

		/* DeviceClose */
		dbg("DeviceClose");
		rc = usb_control_msg(
			serial->dev,
			usb_sndctrlpipe(serial->dev, 0),
			REQ_DEVCLOSE,		/* Request     : setup_packet 2byte */
			REQTYPE_SERIAL,		/* RequestType : setup_packet 1byte  */
			0x0,			/* Value       : setup_packet 3byte  */
			0x0,			/* Index       : setup_packet 4byte  */
			NULL,
			0x0,			/* length      : setup_packet 6-7byt */
			WDR_TIMEOUT);
		if (rc < 0) {
			err("DeviceClose (error = %d)", rc);
		}
	}
	
} /* suntac_ucp_close */


/*
 *	write (use urb_pool)
 */
static int suntac_ucp_write (struct usb_serial_port *port, int from_user,
			   const unsigned char *buf, int count)
{
	struct usb_serial *serial = port->serial;
	unsigned long flags;
	int bytes_sent, size;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;

	dbg("%s - port %d", __FUNCTION__, port->number);

	if (count == 0) {
		dbg("%s - write request of 0 bytes", __FUNCTION__ );
		return (0);
	}

	/* only do something if we have a bulk out endpoint */
	if (!serial->num_bulk_out)
		return (0);
	
	//dbg("write %d [byte] dev:%02x, line:%02x", count, priv->ucp.dev_state, priv->ucp.line_state );

	bytes_sent = 0;
	while ( (count > 0) && (priv->urb_free_cnt > 0) ) {
		
		spin_lock_irqsave (&priv->write_urb_pool_lock, flags);
		
		size = (count > URB_TRANSFER_BUFFER_SIZE) ? URB_TRANSFER_BUFFER_SIZE : count;
		
		usb_serial_debug_data (__FILE__, __FUNCTION__, size, buf);
		
		if (from_user) {
			copy_from_user(priv->writebuf[priv->urb_write_index], buf, size);
		} else {
			memcpy (priv->writebuf[priv->urb_write_index], buf, size);
		}
		priv->writelen[priv->urb_write_index] = size;
		
		bytes_sent += size;

		buf += size;
		count -= size;

		/* Update urb_write_index */
		priv->urb_write_index++;
		if ( priv->urb_write_index >= NUM_URBS )
			priv->urb_write_index = 0;

		/* Decrement of free URB count */
		priv->urb_free_cnt--;

		spin_unlock_irqrestore (&priv->write_urb_pool_lock, flags);
	}
	
	return bytes_sent;
} /* suntac_ucp_write */


/*
 *	write callback
 */
static void suntac_ucp_write_bulk_callback (struct urb *urb)
{
	struct usb_serial_port *port = (struct usb_serial_port *)urb->context;
	struct usb_serial *serial = port->serial;

	dbg("%s - port %d", __FUNCTION__, port->number);
	
	if (!serial) {
		dbg("%s - bad serial pointer, exiting", __FUNCTION__ );
		return;
	}

	if (urb->status) {
		dbg("%s - nonzero write bulk status received: %d", __FUNCTION__, urb->status );
		return;
	}

	/* from generic_write_bulk_callback */
	queue_task(&port->tqueue, &tq_immediate);
	mark_bh(IMMEDIATE_BH);

	return;
} /* suntac_ucp_write_bulk_callback */


static int suntac_ucp_write_room (struct usb_serial_port *port)
{
	unsigned long flags;
	int room = 0;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;

	dbg("%s - port %d", __FUNCTION__, port->number);

	spin_lock_irqsave( &priv->write_urb_pool_lock, flags );

	room = priv->urb_free_cnt * URB_TRANSFER_BUFFER_SIZE;

	spin_unlock_irqrestore( &priv->write_urb_pool_lock, flags );

	dbg("%s - returns %d", __FUNCTION__, room);
	return (room);
}


static int suntac_ucp_chars_in_buffer (struct usb_serial_port *port)
{
	unsigned long flags;
	int chars = 0;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;

	dbg("%s - port %d", __FUNCTION__, port->number);

	spin_lock_irqsave( &priv->write_urb_pool_lock, flags );

	chars = (NUM_URBS - priv->urb_free_cnt) * URB_TRANSFER_BUFFER_SIZE;

	spin_unlock_irqrestore( &priv->write_urb_pool_lock, flags );

	dbg("%s - returns %d", __FUNCTION__, chars);
	return (chars);
}


/*
 *	Interrupt Transfer callback (receive)
 *
 *	status read and submit_urb for data transfer (send, receive both).
 */
static void suntac_ucp_read_int_callback (struct urb *urb)
{
	struct usb_serial_port *port = (struct usb_serial_port *)urb->context;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;
	struct usb_serial *serial = port->serial;
	unsigned char *data = urb->transfer_buffer;
	int		rcvreq;
	int		rcvlen;
	unsigned long	flags;
	int		result;


	dbg("%s - port %d", __FUNCTION__, port->number);

	/* The urb might have been killed. */
	if (urb->status) {
		dbg("%s - nonzero read bulk status received: %d", __FUNCTION__, urb->status);
		return;
	}
	if (!serial) {
		dbg("%s - bad serial pointer, exiting", __FUNCTION__);
		return;
	}
	
	usb_serial_debug_data (__FILE__, __FUNCTION__, urb->actual_length, data);

	/* status changed */
	if ( (priv->ucp.dev_state != data[0]) || (priv->ucp.line_state != data[1]) ) {
		dbg("status = dev:%02x, line:%02x", data[0], data[1] );
	}

	/*
	 * The interrupt-in pipe signals exceptional conditions (modem line
	 * signal changes and errors). 
	 */
	priv->ucp.dev_state  = data[0];
	priv->ucp.line_state = data[1];

	/* Record Control Line states */
	suntac_ucp_msr_to_state(&priv->control_state, priv->ucp.line_state);

	/* Can send ? */
	if ( ((priv->ucp.dev_state >> 2) & 0x03) > 0 ) {
		/* Have Send Data ? */
		if ( (priv->urb_free_cnt < NUM_URBS) && (port->write_urb->status != -EINPROGRESS) ) {

			dbg("%d - SEND_BULK_URB", priv->urb_read_index );

			spin_lock_irqsave( &priv->write_urb_pool_lock, flags );

			/* transfer data copy */
			memcpy( port->write_urb->transfer_buffer, priv->writebuf[priv->urb_read_index], priv->writelen[priv->urb_read_index] );

			/* set up our urb */
			usb_fill_bulk_urb(
				port->write_urb,
				serial->dev,
				usb_sndbulkpipe(serial->dev, port->bulk_out_endpointAddress),
				port->write_urb->transfer_buffer,
				priv->writelen[priv->urb_read_index],
				suntac_ucp_write_bulk_callback,
				port);

			/* send the data out the bulk port */
			result = usb_submit_urb( port->write_urb );

			if ( result ) {
				err("%s - failed submitting write urb, error %d", __FUNCTION__, result );
			}else{
				priv->urb_read_index++;
				if ( priv->urb_read_index >= NUM_URBS )
					priv->urb_read_index = 0;

				priv->urb_free_cnt++;
				if ( priv->urb_free_cnt > NUM_URBS )
					priv->urb_free_cnt = NUM_URBS;
			}
			spin_unlock_irqrestore( &priv->write_urb_pool_lock, flags );
		}
	}

	/* Data Arrived ? */
	if ( (rcvreq = priv->ucp.dev_state & 0x03) > 0 ) {
		rcvlen = 64 * rcvreq;	/* 64, 128, 192 */

		if ( rcvlen > 64 ) {
			dbg("balk_read_start [%d] bytes", rcvlen );
		}

		/* receive data alived ? */
		if ( port->read_urb->status != -EINPROGRESS ) {
			/* Start reading from the device */
			port->read_urb->dev = port->serial->dev;
			usb_fill_bulk_urb(port->read_urb, serial->dev,
				usb_rcvbulkpipe(serial->dev, port->bulk_in_endpointAddress),
				port->read_urb->transfer_buffer,
				rcvlen /* port->read_urb->transfer_buffer_length */,
				suntac_ucp_read_bulk_callback,
				port);

			if (usb_submit_urb(port->read_urb))
				err("usb_submit_urb(read bulk) failed (;_;)");
		}
	}

	/* INT urbs are automatically re-submitted */
} /* suntac_ucp_read_int_callback */


/*
 *	one-shot bulk transfer (receive)
 */
static void suntac_ucp_read_bulk_callback (struct urb *urb)
{
	struct usb_serial_port *port = (struct usb_serial_port *)urb->context;
	struct usb_serial *serial = get_usb_serial (port, __FUNCTION__);
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;
	struct tty_struct *tty;
	unsigned char *data = urb->transfer_buffer;
	int	i;
	char	error_flag = 0;

	if (port_paranoia_check (port, __FUNCTION__))
		return;

	dbg("%s - port %d", __FUNCTION__, port->number);

	if (!serial) {
		dbg("%s - bad serial pointer, exiting", __FUNCTION__ );
		return;
	}

	if (urb->status) {
		dbg("%s - nonzero read bulk status received: %d", __FUNCTION__, urb->status);
		return;
	}

	usb_serial_debug_data (__FILE__, __FUNCTION__, urb->actual_length, data);

	/*
		Check BREAK Condition and Set error_flag (TTY_BREAK).
	*/
	if ( priv->ucp.line_state & 0x80 ) {
		error_flag = TTY_BREAK;
		dbg("BREAK!");
	}

	tty = port->tty;
	if ( tty && urb->actual_length ) {
		for (i = 0; i < urb->actual_length ; ++i) {
			/* if we insert more than TTY_FLIPBUF_SIZE characters, we drop them. */
			if(tty->flip.count >= TTY_FLIPBUF_SIZE) {
				tty_flip_buffer_push(tty);
			}
			/* this doesn't actually push the data through unless tty->low_latency is set */
			tty_insert_flip_char(tty, data[i], error_flag);
		}
		tty_flip_buffer_push(tty);
	}

} /* suntac_ucp_read_bulk_callback */


/*
 *	Serial Setting
 */
static void suntac_ucp_set_termios (struct usb_serial_port *port,
				  struct termios *old_termios)
{
	struct usb_serial *serial = port->serial;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;
	unsigned int iflag = port->tty->termios->c_iflag;
	unsigned int old_iflag = old_termios->c_iflag;
	unsigned int cflag = port->tty->termios->c_cflag;
	unsigned int old_cflag = old_termios->c_cflag;

	
	/*
	 * Update baud rate
	 */
	if( (cflag & CBAUD) != (old_cflag & CBAUD) ) {
		/* reassert DTR and (maybe) RTS on transition from B0 */
		if( (old_cflag & CBAUD) == B0 ) {
			dbg("%s: baud was B0", __FUNCTION__);
			priv->control_state |= TIOCM_DTR;
			/* don't set RTS if using hardware flow control */
			if (!(old_cflag & CRTSCTS)) {
				priv->control_state |= TIOCM_RTS;
			}
			suntac_ucp_mdmctl(serial, priv->control_state);
		}
		
		switch(cflag & CBAUD) {
		case B0: /* handled below */
			break;
		case B110: suntac_ucp_set_baud_rate(serial, 110);
			break;
		case B300: suntac_ucp_set_baud_rate(serial, 300);
			break;
		case B600: suntac_ucp_set_baud_rate(serial, 600);
			break;
		case B1200: suntac_ucp_set_baud_rate(serial, 1200);
			break;
		case B2400: suntac_ucp_set_baud_rate(serial, 2400);
			break;
		case B4800: suntac_ucp_set_baud_rate(serial, 4800);
			break;
		case B9600: suntac_ucp_set_baud_rate(serial, 9600);
			break;
		case B19200: suntac_ucp_set_baud_rate(serial, 19200);
			break;
		case B38400: suntac_ucp_set_baud_rate(serial, 38400);
			break;
		case B57600: suntac_ucp_set_baud_rate(serial, 57600);
			break;
		case B115200: suntac_ucp_set_baud_rate(serial, 115200);
			break;
		case B230400: suntac_ucp_set_baud_rate(serial, 230400);
			break;
		case B460800: suntac_ucp_set_baud_rate(serial, 460800);
			break;
		case B921600: suntac_ucp_set_baud_rate(serial, 921600);
			break;
		default: err("SUNTAC U-Cable P : unsupported baudrate request, using default of 115200");
			suntac_ucp_set_baud_rate(serial, 115200); break;
		}
		if ((cflag & CBAUD) == B0 ) {
			dbg("%s: baud is B0", __FUNCTION__ );
			/* Drop RTS and DTR */
			priv->control_state &= ~(TIOCM_DTR | TIOCM_RTS);
			suntac_ucp_mdmctl(serial, priv->control_state);
		}
	}

	/*
	 *  Update Data Transfer Protocol
	 */
	if ((cflag & (PARENB|PARODD)) != (old_cflag & (PARENB|PARODD))
		|| (cflag & CSIZE) != (old_cflag & CSIZE)
		|| (cflag & CSTOPB) != (old_cflag & CSTOPB) ) {
		
		priv->databits = 0;

		/* set the parity */
		if (cflag & PARENB)
			priv->databits |= (cflag & PARODD) ?
				SUNTAC_UCP_PARITY_USE | SUNTAC_UCP_PARITY_ODD :
				SUNTAC_UCP_PARITY_USE | SUNTAC_UCP_PARITY_EVEN; 
		else
			priv->databits |= SUNTAC_UCP_PARITY_NOUSE;

		/* set the number of data bits */
		switch (cflag & CSIZE) {
		case CS7:
			priv->databits |= SUNTAC_UCP_DATA_BITS_7; break;
		case CS8:
			priv->databits |= SUNTAC_UCP_DATA_BITS_8; break;
		default:
			err("CSIZE was not CS7-CS8, using default of 8");
			priv->databits |= SUNTAC_UCP_DATA_BITS_8;
			break;
		}

		/* set the number of stop bits (but, U-Cable P2 is not use this). */
		priv->databits |= (cflag & CSTOPB) ? 0 : 0;

		/* set databits */
		suntac_ucp_set_protocol(serial, priv->databits);
	}
	
	if( (iflag & IXOFF) != (old_iflag & IXOFF)
		|| (iflag & IXON) != (old_iflag & IXON)
		||  (cflag & CRTSCTS) != (old_cflag & CRTSCTS) ) {
		
		/* Drop DTR/RTS if no flow control otherwise assert */
		if ((iflag & IXOFF) || (iflag & IXON) || (cflag & CRTSCTS) )
			priv->control_state |= TIOCM_DTR | TIOCM_RTS;
		else
			priv->control_state &= ~(TIOCM_DTR | TIOCM_RTS);
		suntac_ucp_mdmctl(serial, priv->control_state);
	}
} /* suntac_ucp_set_termios */


/*
 *	Break control.
 */
static void suntac_ucp_break_ctl( struct usb_serial_port *port, int break_state )
{
	struct usb_serial *serial = port->serial;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;

	dbg("%s - state=%d", __FUNCTION__, break_state);

	if (break_state)
		priv->ctrlline |= 0x80;
	else
		priv->ctrlline &= ~0x80;

	suntac_ucp_mdmctrl_direct(serial, priv->ctrlline);
} /* suntac_ucp_break_ctl */


/*
 *	IOCTL
 */
static int suntac_ucp_ioctl (struct usb_serial_port *port, struct file * file,
			   unsigned int cmd, unsigned long arg)
{
	struct usb_serial *serial = port->serial;
	struct suntac_ucp_private *priv = (struct suntac_ucp_private *)port->private;
	int ret, mask;

	
	dbg("%s - cmd=0x%x", __FUNCTION__, cmd);

	/* Based on code from acm.c and others */
	switch (cmd) {
	case TIOCMGET:	// return now control_state
		return put_user(priv->control_state, (unsigned long *) arg);
		break;

	case TIOCMSET: /* Turns on and off the lines as specified by the mask */
	case TIOCMBIS: /* turns on (Sets) the lines as specified by the mask */
	case TIOCMBIC: /* turns off (Clears) the lines as specified by the mask */
		if ((ret = get_user(mask, (unsigned long *) arg))) return ret;

		if ((cmd == TIOCMSET) || (mask & TIOCM_RTS)) {
			/* RTS needs set */
			if( ((cmd == TIOCMSET) && (mask & TIOCM_RTS)) ||
			    (cmd == TIOCMBIS) )
				priv->control_state |=  TIOCM_RTS;
			else
				priv->control_state &= ~TIOCM_RTS;
		}

		if ((cmd == TIOCMSET) || (mask & TIOCM_DTR)) {
			/* DTR needs set */
			if( ((cmd == TIOCMSET) && (mask & TIOCM_DTR)) ||
			    (cmd == TIOCMBIS) )
				priv->control_state |=  TIOCM_DTR;
			else
				priv->control_state &= ~TIOCM_DTR;
		}
		suntac_ucp_mdmctl(serial, priv->control_state);
		break;
					
	case TIOCMIWAIT:
		/* wait for any of the 4 modem inputs (DCD,RI,DSR,CTS)*/
		/* TODO */
		return( 0 );

	case TIOCGICOUNT:
		/* return count of modemline transitions */
		/* TODO */
		return 0;

	default:
		dbg("%s: arg not supported - 0x%04x", __FUNCTION__, cmd);
		return(-ENOIOCTLCMD);
		break;
	}
	return 0;
} /* suntac_ucp_ioctl */


static void suntac_ucp_throttle (struct usb_serial_port *port)
{
	dbg("%s - port %d", __FUNCTION__, port->number);
	usb_unlink_urb(port->interrupt_in_urb);
}


static void suntac_ucp_unthrottle (struct usb_serial_port *port)
{
	int	result;

	dbg("%s - port %d", __FUNCTION__, port->number);

	port->interrupt_in_urb->dev = port->serial->dev;
	result = usb_submit_urb(port->interrupt_in_urb);
	if ( result )
		err("%s - failed submitting interrupt urb, error %d", __FUNCTION__, result);
}


/*
 *	module insert
 */
static int __init suntac_ucp_init (void)
{
	usb_serial_register (&suntac_ucp_device);
	info(DRIVER_DESC " " DRIVER_VERSION);
	return 0;
}


/*
 *	module remove
 */
static void __exit suntac_ucp_exit (void)
{
	usb_serial_deregister (&suntac_ucp_device);
}

module_init(suntac_ucp_init);
module_exit(suntac_ucp_exit);

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug enabled or not");

